import typing
import warnings

import pydantic

from ..bids_models import DatasetDescription


def get_bids_dataset_description(dandiset) -> DatasetDescription:

    valid_or_raw: typing.Literal["raw", "valid"] = "valid"
    try:
        metadata = dandiset.get_metadata()
    except pydantic.ValidationError:
        # TODO
        # raw_dandiset_metadata = dandiset.get_raw_metadata()
        valid_or_raw = "raw"

    if valid_or_raw == "valid":
        dataset_description = _get_dataset_description_from_valid_dandiset_metadata(metadata=metadata)
    else:
        message = "Handling invalid Dandiset metadata is not yet implemented."
        raise NotImplementedError(message)

    return dataset_description


def _get_dataset_description_from_valid_dandiset_metadata(metadata: dict) -> DatasetDescription:
    dataset_description_kwargs = dict()

    dataset_description_kwargs["Name"] = metadata.name
    dataset_description_kwargs["BIDSVersion"] = "1.10"

    bids_rrid = "RRID:SCR_016124"
    if any(data_standard.identifier == bids_rrid for data_standard in metadata.assetsSummary.dataStandard):
        message = (
            "Dandiset already contains BIDS content. If only a partial conversion is desired, "
            "please raise an issue on https://github.com/con/nwb2bids/issues/new to discuss the use case."
        )
        warnings.warn(message=message, category=RuntimeWarning)
        return None

    if metadata.description is not None:
        dataset_description_kwargs["Description"] = metadata.description
    if metadata.contributor is not None:
        dataset_description_kwargs["Authors"] = [
            contributor.name
            for contributor in metadata.contributor
            for role in contributor.roleName
            if role.value == "dcite:Author"
        ]
    if metadata.license is not None:
        dataset_description_kwargs["License"] = metadata.license[0].value.split(":")[1]

    dataset_description = DatasetDescription(**dataset_description_kwargs)
    return dataset_description



import pydantic


class BaseMetadataModel(pydantic.BaseModel):
    """
    Base Pydantic model for all metadata handled by `nwb2bids`.
    """

    model_config = pydantic.ConfigDict(
        validate_assignment=True,  # Re-validate model on mutation
        extra="allow",  # Allow additional custom fields
    )



import pathlib

import h5py
import pydantic
import pynwb
import typing_extensions

from ._base_metadata_model import BaseMetadataModel
from ._channels import ChannelTable
from ._electrodes import ElectrodeTable
from ._events import Events
from ._participant import Participant
from ._probes import ProbeTable
from .._tools import cache_read_nwb


class BidsSessionMetadata(BaseMetadataModel):
    """
    Schema for the metadata of a single BIDS session.
    """

    session_id: str = pydantic.Field(
        description="A unique session identifier.",
        pattern=r"^[^_]+$",  # No underscores allowed
    )
    participant: Participant = pydantic.Field(description="Metadata about a participant used in this experiment.")
    events: Events | None = pydantic.Field(
        description="Timing data and metadata regarding events that occur during this experiment.", default=None
    )
    probe_table: ProbeTable | None = None
    channel_table: ChannelTable | None = None
    electrode_table: ElectrodeTable | None = None

    @classmethod
    @pydantic.validate_call
    def from_nwbfile_paths(
        cls, nwbfile_paths: list[pydantic.FilePath] | list[pydantic.HttpUrl]
    ) -> typing_extensions.Self:
        # Differentiate local path from URL
        if isinstance(nwbfile_paths[0], pathlib.Path):
            nwbfiles = [cache_read_nwb(nwbfile_path) for nwbfile_path in nwbfile_paths]
        else:
            nwbfiles = [_stream_nwb(url=url) for url in nwbfile_paths]

        session_ids = list({nwbfile.session_id for nwbfile in nwbfiles})
        if len(session_ids) > 1:
            message = "Multiple differing session IDs found - please check how this method was called."
            raise ValueError(message)
        session_id = session_ids[0]

        participant = Participant.from_nwbfiles(nwbfiles=nwbfiles)
        events = Events.from_nwbfiles(nwbfiles=nwbfiles)
        probe_table = ProbeTable.from_nwbfiles(nwbfiles=nwbfiles)
        channel_table = ChannelTable.from_nwbfiles(nwbfiles=nwbfiles)
        electrode_table = ElectrodeTable.from_nwbfiles(nwbfiles=nwbfiles)

        dictionary = {
            "session_id": session_id,
            "participant": participant,
        }
        if events is not None:
            dictionary["events"] = events
        if probe_table is not None:
            dictionary["probe_table"] = probe_table
        if channel_table is not None:
            dictionary["channel_table"] = channel_table
        if electrode_table is not None:
            dictionary["electrode_table"] = electrode_table

        session_metadata = BidsSessionMetadata(**dictionary)
        return session_metadata


def _stream_nwb(url: pydantic.HttpUrl) -> pynwb.NWBFile:
    """
    Stream an NWB file from a URL using remfile.

    Parameters
    ----------
    url : pydantic.HttpUrl
        The URL of the NWB file to stream.

    Returns
    -------
    pynwb.NWBFile
        The streamed NWB file.
    """
    import remfile

    rem_file = remfile.File(url=str(url))

    try:
        h5py_file = h5py.File(name=rem_file, mode="r")
    except Exception as exception:
        message = (
            f"\nFailed to open NWB file from URL {url}: {exception}\n\n"
            "Possible that backend is not supported.\n"
            "Please raise an issue on https://github.com/con/nwb2bids/issues/new to discuss."
        )
        raise ValueError(message)

    file_io = pynwb.NWBHDF5IO(file=h5py_file, mode="r")
    nwbfile = file_io.read()
    return nwbfile



"""Configuration file for the Sphinx documentation builder."""

project = "nwb2bids"
copyright = "2025, Cody Baker"
author = "Cody Baker"

extensions = []

exclude_patterns = []

html_theme = "pydata_sphinx_theme"

html_theme_options = {
    "github_url": "https://github.com/con/nwb2bids",
    "use_edit_page_button": True,
    "show_toc_level": 2,
}

html_context = {
    "github_user": "con",
    "github_repo": "nwb2bids",
    "github_version": "main",
    "doc_path": "docs/source",
}



import re
import warnings

import pydantic
import pynwb
import typing_extensions

from ._base_metadata_model import BaseMetadataModel
from ._model_globals import _ALLOWED_SEXES, _SPECIES_REGEX


class Participant(BaseMetadataModel):
    participant_id: str = pydantic.Field(
        description="A unique identifier for this participant.",
        pattern=r"^[^_]+$",  # No underscores allowed
    )
    species: str | None = pydantic.Field(
        description=(
            "The species should be the proper Latin binomial species name from the NCBI Taxonomy "
            "(for example, Mus musculus)."
        ),
        pattern=_SPECIES_REGEX,
        default=None,
    )
    sex: str | None = pydantic.Field(
        description=(
            'String value indicating phenotypical sex, one of "male", "female", "other".\n'
            '\tFor "male", use one of these values: male, m, M, MALE, Male.\n'
            '\tFor "female", use one of these values: female, f, F, FEMALE, Female.\n'
            '\tFor "other", use one of these values: other, o, O, OTHER, Other.'
        ),
        default=None,
    )
    strain: str | None = pydantic.Field(
        description=(
            "For species other than Homo sapiens, the string value indicating the strain of the species "
            "(for example, C57BL/6J)."
        ),
        default=None,
    )

    @classmethod
    @pydantic.validate_call
    def from_nwbfiles(cls, nwbfiles: list[pydantic.InstanceOf[pynwb.NWBFile]]) -> typing_extensions.Self:
        """
        Extracts participant metadata from the in-memory NWBFile objects.
        """
        if len(nwbfiles) > 1:
            message = "Conversion of multiple NWB files per session is not yet supported."
            raise NotImplementedError(message)

        nwbfile = nwbfiles[0]
        participant_id = nwbfile.subject.subject_id.replace("_", "-")

        species = nwbfile.subject.species
        if species is not None and not re.match(pattern=_SPECIES_REGEX, string=species):
            message = (
                f"Subject species '{species}' within NWB file is not a proper Latin binomial or NCBI Taxonomy link. "
                "Skipping automated extraction."
            )
            warnings.warn(message=message, stacklevel=2)
            species = None

        sex = nwbfile.subject.sex
        if nwbfile.subject.sex is not None and nwbfile.subject.sex not in _ALLOWED_SEXES:
            message = (
                f"Subject sex '{sex}' within NWB file is not one of the allowed patterns. "
                "Skipping automated extraction."
            )
            warnings.warn(message=message, stacklevel=2)
            sex = None

        subject = cls(
            participant_id=participant_id,
            species=species,
            sex=sex,
            strain=nwbfile.subject.strain,
            # TODO: add more
            # birthday=nwbfile.participant.date_of_birth,
            # age=nwbfile.participant.age
        )
        return subject



_SPECIES_REGEX = r"^([A-Za-z ]+|http://purl\.obolibrary\.org/obo/NCBITaxon_\d+)$"
_ALLOWED_SEXES = {
    "male",
    "m",
    "M",
    "MALE",
    "Male",
    "female",
    "f",
    "F",
    "FEMALE",
    "Female",
    "other",
    "o",
    "O",
    "OTHER",
    "Other",
}



import functools
import pathlib

import pynwb


@functools.cache
def cache_read_nwb(file_path: pathlib.Path) -> pynwb.NWBFile:
    """Cache the read operation per NWB file path to speed up repeated calls."""
    nwbfile = pynwb.read_nwb(path=file_path)
    return nwbfile



import pathlib
import typing

import pydantic

from .._converters._dataset_converter import DatasetConverter


@pydantic.validate_call
def convert_nwb_dataset(
    *,
    nwb_paths: typing.Iterable[pydantic.FilePath | pydantic.DirectoryPath] = pydantic.Field(min_length=1),
    bids_directory: str | pathlib.Path | None = None,
    file_mode: typing.Literal["move", "copy", "symlink", "auto"] = "auto",
    additional_metadata_file_path: pydantic.FilePath | None = None,
) -> None:
    converter = DatasetConverter.from_nwb_paths(
        nwb_paths=nwb_paths,
        additional_metadata_file_path=additional_metadata_file_path,
    )
    converter.extract_metadata()
    converter.convert_to_bids_dataset(bids_directory=bids_directory, file_mode=file_mode)
